use crate::types::{Item, Spell};
use graphql_client::{GraphQLQuery, QueryBody, Response};
use std::error::Error;

// https://studio.apollographql.com/sandbox?endpoint=https%3A%2F%2Fwww.dnd5eapi.co%2Fgraphql

#[derive(GraphQLQuery)]
#[graphql(
    schema_path = "gql/schema.graphql",
    query_path = "gql/GetSpell.graphql",
    response_derives = "Debug"
)]
struct GetSpell;

#[derive(GraphQLQuery)]
#[graphql(
    schema_path = "gql/schema.graphql",
    query_path = "gql/GetItem.graphql",
    response_derives = "Debug"
)]
struct GetItem;

fn perform_request<Q>(
    request_body: QueryBody<Q::Variables>,
) -> Result<Response<Q::ResponseData>, Box<dyn Error>>
where
    Q: GraphQLQuery,
{
    let client = reqwest::blocking::Client::new();

    let res = client
        .post("https://www.dnd5eapi.co/graphql")
        .json(&request_body)
        .send()?;
    let response_body: Response<Q::ResponseData> = res.json()?;

    return Ok(response_body);
}

fn spell_component_to_string(c: &get_spell::SpellComponent) -> String {
    match c {
        get_spell::SpellComponent::M => String::from("M"),
        get_spell::SpellComponent::V => String::from("V"),
        get_spell::SpellComponent::S => String::from("S"),
        get_spell::SpellComponent::Other(s) => String::from(s),
    }
}

fn rarity_to_string(r: &get_item::MagicItemRarity) -> String {
    match r {
        get_item::MagicItemRarity::ARTIFACT => String::from("ARTIFACT"),
        get_item::MagicItemRarity::LEGENDARY => String::from("LEGENDARY"),
        get_item::MagicItemRarity::VERY_RARE => String::from("VERY_RARE"),
        get_item::MagicItemRarity::RARE => String::from("RARE"),
        get_item::MagicItemRarity::UNCOMMON => String::from("UNCOMMON"),
        get_item::MagicItemRarity::COMMON => String::from("COMMON"),
        get_item::MagicItemRarity::VARIES => String::from("VARIES"),
        get_item::MagicItemRarity::Other(s) => String::from(s),
    }
}

pub fn get_spell(index: &str) -> Result<Spell, Box<dyn Error>> {
    let vars = get_spell::Variables {
        index: index.into(),
    };

    let body = GetSpell::build_query(vars);

    let res = perform_request::<GetSpell>(body)?;

    match res.data {
        Some(data) => match data.spell {
            Some(spell) => {
                let rtn = Spell {
                    name: spell.name,
                    level: spell.level,
                    school: spell.school.name,
                    range: spell.range,
                    desc: spell.desc.join(" "),
                    components: match spell.components {
                        Some(components) => {
                            Some(components
                                .iter()
                                .flatten()
                                .map(spell_component_to_string)
                                .collect::<Vec<String>>().join(" ")
                            )
                        }
                        _ => None,
                    },
                    duration: spell.duration,
                    ritual: spell.ritual,
                    casting_time: spell.casting_time,
                };

                return Ok(rtn);
            }

            _ => {
                return Err(Box::from("Spell not found"));
            }
        },
        _ => {
            return Err(Box::from("Spell not found"));
        }
    }
}

pub fn get_item(index: &str) ->  Result<Item, Box<dyn Error>>{
    let vars = get_item::Variables {
        index: index.into(),
    };

    let body = GetItem::build_query(vars);

    let res = perform_request::<GetItem>(body)?;

    match res.data {
        Some(data) => {
            match data.magic_item {
                Some(item) => {
                    let rtn = Item{
                        name: item.name,
                        rarity: rarity_to_string(&item.rarity),
                        desc: item.desc.join(" ")
                    };

                    return Ok(rtn);
                }

                None => {
                    return Err(Box::from("Item not found"));
                }
            }
        }

        None => {
            return Err(Box::from("Item not found"));
        }
    }

}